package DHCPDetect;

# DHCP Detection Library 
#
# version 1.0
#
# This library is a hack of the Net::DHCPClient library
# in cpan. Its quite a bit messy because it was written
# in a crunch to save us some headaches during our
# ResNet Move-in. Please feel free to fix/rewrite
# or improve and send us your changes.
#
# Its tested with Debian Linux (Sarge)
#
# For info see http://oss.oregonstate.edu/roguedetect
#
# This library is made possible by the Oregon State University
# Open Source movement.
#
# Oregon State University
# Open Minds. Open Doors. Open Source.
# http://oregonstate.edu
#
# COPYRIGHT
#
# Based on DHCPClient library by:
#  Joshua Walgenbach, Indiana University, jwalgenb@indiana.edu
#
# Hacked into DHCP Server Detector library by:
#  Alex Schumann, Oregon State University, alex.schumann@oregonstate.edu
#  initial release: Sept 10, 2003
#
# LISCENSE
#
# This program is free software; you can redistribute it and/or modify it under the terms of the 
# GNU General Public License as published by the Free Software Foundation; either version 1, or 
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See either the 
# GNU General Public License or the Artistic License for more details.
#


use strict;

use Carp;
use Net::RawIP qw(:pcap);

use Data::Dumper;
use FileHandle;

use vars qw($AUTOLOAD);

my %fields = (
              interface => undef,
              macaddr => undef,
              src => undef,
              dest => undef,
              op => undef,
              htype => undef,
              hlen => undef,
              hops => undef,
              xid => undef,
              secs => undef,
              flags => undef,
              ciaddr => undef,
              yiaddr => undef,
              siaddr => undef,
              giaddr => undef,
              sname => undef,
              bootfile => undef,
              debug => undef,
              options => undef,
              timeout => undef,
              svr_hwa => undef,
              approved => (),
              );

sub new {
  my $that = shift;
  my $class = ref( $that ) || $that;
  my $self = {
              _permitted => \%fields,
              %fields,
             };

  bless $self, $class;

  if ( @_ ) {
    my %conf = @_;
    while ( my( $k, $v ) = each %conf ) {
      $self->$k( $v );
    }
  }

  return $self;
}

sub AUTOLOAD {
  my $self = shift;
  my $type = ref( $self ) || croak "$self is not an object";
  my $name = $AUTOLOAD;

  if ( @_ ) {
    return $self->{$name} = shift;
  } else {
    return $self->{$name};
  }
}

sub bait {
    my $self = shift;
    my %options = @_;

    $self->ciaddr( 0 );
    $self->yiaddr( 0 );
    $self->siaddr( 0 );
    $self->giaddr( 0 );

    $options{53} = '1';
    $options{55} = '1 28 2 3 15 6 12 40 41 42';

    $self->sendbait( \%options );
}

sub scan {
    my $self = shift;
    my %options = @_;

    $self->ciaddr( 0 );
    $self->yiaddr( 0 );
    $self->siaddr( 0 );
    $self->giaddr( 0 );

    $options{53} = '1';
    $self->doscan( \%options );
}

sub inform {
  my $self = shift;
  my %options = @_;

  $options{53} = '8';

  $self->doit( \%options );
} 

# moved to calling file
#
#sub approvedserver {
#    my $self = shift;
#    my $s_hwa = shift;
#    my $ok = 0;
#
#    foreach $a (@{$self->approved})
#    {
#        if($a eq $s_hwa) 
#        { 
#            if($self->debug > 1)
#            {
#                print "DEBUG $s_hwa is an approved server, ignoring..\n";
#            }
#            return(1);
#        }
#    }
#    if($self->debug > 1)
#    {
#        print "DEBUG: $s_hwa is not on the approved server list!\n";
#    }
#    return(0);
#}

sub sendbait {
  my $self = shift;
  my $o = shift; 
  my %options = %$o;

  my $macaddr = $self->macaddr;
  my $interface = $self->interface;

  my $op = 1; 
  my $srcip = 0;
  my $htype = $self->htype || 1;
  my $hlen = $self->hlen || 6;
  my $ciaddr = $self->client_ip() || '0.0.0.0';
  my $yiaddr = $self->your_ip() || '0.0.0.0';
  my $siaddr = $self->server_ip() || '0.0.0.0';
  my $giaddr = $self->relay_ip() || '0.0.0.0';
  my $sname = $self->sname || 0;
  my $file = $self->bootfile || 0;

  my $timeout = $self->timeout || 10;

  my $dstmac;
  my $dstip;

  if($self->debug > 1)
  {
    print("DEBUG - Sending a bait packet\n");
  }

  if ( $self->src ) {
    $dstmac = net2mac( $self->src );
  } else {
    $dstmac = 'ff:ff:ff:ff:ff:ff';
  }

  if ( $self->siaddr ) {
    $dstip = $self->server_ip();
  } else {
    $dstip = '255.255.255.255';
  }

  my $srcip = $self->your_ip();

  my $xid = sprintf "%0.8d", int( rand 99999999 );

  my $data = $self->encode( op => $op, htype => $htype, hlen => $hlen, 
                            hops => 0, xid => $xid, secs => 0, flags => 0, 
                            ciaddr => $ciaddr, yiaddr => $yiaddr, 
                            siaddr => $siaddr, giaddr => $giaddr, 
                            chaddr => $macaddr, sname => $sname, 
                            file => $file, options => \%options );

  my $p = new Net::RawIP( {udp => {}} );

  $p->ethnew( $interface );
  $p->ethset( source => $macaddr, dest => $dstmac);

  $p->set( {ip => {saddr => $srcip, daddr => $dstip},
            udp => {source => 68, dest => 67, data => $data}} );

#  my $filter = "ether dst $macaddr and dst port 68";

  my $a = new Net::RawIP;
  my @f;

  my $done = 0;

#  STDERR->close;

#  my $pcap = $a->pcapinit( $interface, $filter, 1500, 500 );

#  STDERR->fdopen( 2, "w+" );


#  $SIG{ALRM} = sub { die "timeout" };

  while ( ! $done ) {

    if ( $self->debug > 2) {

      my $op53 = $options{'53'};

      print "XMIT:\n";
      printf
        "xid %u:\tsaddr=%s daddr=%s\n\t\tsrc=%s dest=%s op=%d\n",
          $xid, $srcip, $dstip, $macaddr, $dstmac, $op53;
    }

    $p->ethsend;

    if ( $options{'53'} == 7 ) {
      return 1;
    }
  return 1;
}



sub doscan {
  my $self = shift;
  my $o = shift; 
  my %options = %$o;

  my $macaddr = $self->macaddr;
  my $interface = $self->interface;

  my $op = 1; 
  my $svrip = '';
  my $htype = $self->htype || 1;
  my $hlen = $self->hlen || 6;
  my $ciaddr = $self->client_ip() || '0.0.0.0';
  my $yiaddr = $self->your_ip() || '0.0.0.0';
  my $siaddr = $self->server_ip() || '0.0.0.0';
  my $giaddr = $self->relay_ip() || '0.0.0.0';
  my $sname = $self->sname || 0;
  my $file = $self->bootfile || 0;

  my $timeout = $self->timeout || 10;

  my $dstmac;
  my $dstip;

  if ( $self->src ) {
    $dstmac = net2mac( $self->src );
  } else {
    $dstmac = 'ff:ff:ff:ff:ff:ff';
  }

  if ( $self->siaddr ) {
    $dstip = $self->server_ip();
  } else {
    $dstip = '255.255.255.255';
  }

  my $srcip = $self->your_ip();

#  my $xid = sprintf "%0.8d", int( rand 99999999 );

#  my $data = $self->encode( op => $op, htype => $htype, hlen => $hlen, 
#                           hops => 0, xid => $xid, secs => 0, flags => 0, 
##                          ciaddr => $ciaddr, yiaddr => $yiaddr, 
#                           siaddr => $siaddr, giaddr => $giaddr, 
#                           chaddr => $macaddr, sname => $sname, 
#                           file => $file, options => \%options );

  my $p = new Net::RawIP( {udp => {}} );

  $p->ethnew( $interface );
  $p->ethset( source => $macaddr, dest => $dstmac);

  $p->set( {ip => {saddr => $srcip, daddr => $dstip},
            udp => {source => 68, dest => 67, data => $data}} );

  #my $filter = "ether dst $macaddr and dst port 68";
  my $filter = "dst port 68";

  my $a = new Net::RawIP;
  my @f;

  my $done = 0;

#  STDERR->close;

  if($self->debug >4)
  {
     print "Starting a pcap: $interface, $filter\n";
  }
  my $pcap = $a->pcapinit( $interface, $filter, 1500, 500 );
#  STDERR->fdopen( 2, "w+" );

    # While the server ip is approved or empty, restart the sniffer..
    #while ($svrip eq ''  or $self->approvedserver(net2mac($self->src)) or $self->reply ne 'offer') 
#    while ($svrip eq '') 
#    {
        if($self->debug >3)
        {
           print "Listener is waiting for a packet...\n";
        }
        # Listen for the next packet..
        
        # call the sniffing loop, with the following code as an anonymous callback..
        loop $pcap, 1, sub 
        { 
          if($self->debug >3)
          {
             print "Listener GOT a packet...\n";
          }
          my $time = timem();
          $p->bset( substr( $_[2], 14 ) );
        
          $self->dest( substr( $_[2], 0, 6) );
          $self->src( substr( $_[2], 6, 6) );
        
          my @g = $p->get( {ip => [qw(ttl saddr daddr)],
                            udp => [qw(data)]} );

          $self->decode( $g[3] );
          #$self->srcmac = net2mac($self->src);
          #$self->dstmac = net2mac($self->dest);

          $svrip = $self->server_ip;
          $done = 1;

          if ( $self->debug > 3) 
          {
        
            my $x = $self->options();
            my %options = %$x;

            my $op53 = $options{'53'};

            print "CHILD - RCVD:\n";
            printf 
              "CHILD - xid %lx:\n    %lu\tttl=%u saddr=%s daddr=%s\n\t\tsrc=%s dest=%s op=%d\n",
                $self->xid, $xid, $g[0], ip2dot( $g[1] ), ip2dot( $g[2] ),
                  net2mac( $self->src ), net2mac( $self->dest ), $op53;
            print "CHILD - (".$self->reply . " " .$self->server_ip.")\n";
          }
          my $svr_hwa = net2mac( $self->src );
          if ( $self->debug > 5 ) 
          {
            printf "%u %x\n", $xid, $self->xid;
          }
        
        }, \@f;  # End of anony sub passed to loop..
#    }

    if ( $@ ) {
      if ( $@ =~ /timeout/ ) {
        if ( $self->debug > 1 ) {
          print "CHILD - TIMEOUT\n";
        }
        return 0;
      } else {
        print "weird stuff at the bottom.\n";
        die;
        }
    }
  }

  my $x = $self->options();

  my %options = %$x;

  if ( $self->debug > 2 ) {
    print "Returning from doscan()\n";
  }

  return 1;
}

sub decode {
  my $self = shift;
  my $data = shift;
  my @fields = ('op', 'htype', 'hlen', 'hops', 'xid', 'secs', 'flags',
                'ciaddr', 'yiaddr', 'siaddr', 'giaddr');
  my %options; 

  my $bootp = substr $data, 0, 239;
  my $opts = substr $data, 240;
  my @bootp = unpack "H2 H2 H2 H2 H8 H4 H4 H8 H8 H8 H8 H32 H128 H256", $bootp;

  for ( my $i;  $i <= 10; $i++ ) { $bootp[$i] = hex( $bootp[$i] ); }

  for ( my $i; $i <= $#bootp; $i++ ) {
    my $method = $fields[$i];
    $self->$method( $bootp[$i] );
  }

  my @opts = unpack "C" x ( length( $data ) - 240 ), $opts;

  for ( my $i = 0; $i <= $#opts; $i++ ) {
    
    my $opt = $opts[$i++];
    my $len = $opts[$i++];
    my $offset = $len + $i - 1;
    my $string = "";

    for ( my $q = $i; $q <= $offset; $q++ ) {
      if ( $string ) {
        $string = sprintf "%s %d", $string, $opts[$q];
      } else {
        $string = sprintf "%d", $opts[$q];
      }
    }

    $options{$opt} = $string;

    $i = $i + $len - 1;
  }

  $self->options( \%options );
}

sub encode {
  my $self = shift;
  my %c = @_;
  
  my $magic = pack "C4", 99, 130, 83, 99;

  my @ciaddr = split /\./, $c{'ciaddr'};
  my @yiaddr = split /\./, $c{'yiaddr'};
  my @siaddr = split /\./, $c{'siaddr'};
  my @giaddr = split /\./, $c{'giaddr'};
  
  my @chaddr = mac2net( $c{'chaddr'} );

  my $data = pack "C4 H8 H4 H4 C4 C4 C4 C4 C16 H128 H256", 
  $c{'op'}, $c{'htype'}, $c{'hlen'}, $c{'hops'}, $c{'xid'}, $c{'secs'},
  $c{'flags'}, @ciaddr, @yiaddr, @siaddr, @giaddr, @chaddr, $c{'sname'}, 
  $c{'file'};

  $data = join '', $data, $magic;

  my $o = $c{'options'}; 
  my %options = %$o;

  foreach my $key ( keys %options ) {
    my @p = split / /, $options{$key};
    map { $_ = hex( $_ ); } @p;

    my $format = sprintf "C%d", $#p+3;

    my $options = pack $format, $key, $#p+1, @p;

    $data = join '', $data, $options;
  }

  my $end = pack "C2", 255, 0;

  $data = join '', $data, $end;

  return $data;
}

sub net2mac {
  return sprintf( "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", unpack( "C6", shift ) );
}

sub mac2net {
  my @a = split /:/, shift;

  for ( 1..10 ) {
    push @a, '0';
  }

  map { $_ = hex($_); } @a;
    
  return @a;
}

sub ip2dot {
  return sprintf( "%u.%u.%u.%u", unpack( "C4", pack( "N", shift ) ) );
}

sub dot2ip {
  return unpack( "N", pack( "C4", split( /\./, shift ) ) );
}

sub client_ip {
  my $self = shift;

  return ip2dot( $self->ciaddr );
}

sub reply {
  my $self = shift;

  my @ops = ( '', 'discover', 'offer', 'request', 'decline', 'ack', 'nak',
              'release', 'inform' );

  my $x = $self->options;
  my %options = %$x;

  return ( $ops[$options{'53'}] );
}

sub your_ip {
  my $self = shift;

  return ip2dot( $self->yiaddr );
}

sub svr_hwa {
  my $self = shift;
  return net2mac($self->src);
}

sub server_ip {
  my $self = shift;

  return ip2dot( $self->siaddr );
}

sub relay_ip {
  my $self = shift;

  return ip2dot( $self->giaddr );
}

sub domain_name {
  my $self = shift;

  my $o = $self->options;
  my %options = %$o;

  my @l = split / /, $options{15};

  return( pack "C*", @l );
}

sub server_lookup {
  my $self = shift;
  my $option = shift;

  my $o = $self->options;
  my %options = %$o;

  my @val = split / /, $options{$option};

  my @servers;

  for ( my $i; $i <= $#val; $i++ ) {
    push @servers, sprintf "%d.%d.%d.%d", 
           $val[$i++], $val[$i++], $val[$i++], $val[$i];
  }

  return @servers; 
}
  
sub domain_name_server {
  my $self = shift;

  return $self->server_lookup(6);
}

sub server_identifier {
  my $self = shift;

  return $self->server_lookup(54);
}

sub router {
  my $self = shift;

  return $self->server_lookup(3);
}

sub subnet_mask {
  my $self = shift;
  
  return $self->server_lookup(1);
}

sub ntp_server {
  my $self = shift;
  
  return $self->server_lookup(42);
}

sub time_server {
  my $self = shift;

  return $self->server_lookup(4);
}

sub netbios_nameserver {
  my $self = shift;
  
  return $self->server_lookup(44);
}

sub broadcast {
  my $self = shift;

  return $self->server_lookup(28);
}

1;

